//
//  NSString+JapaneseExtension.swift
//  migemodoki
//
//  Created by narita on 2017/10/30.
//  Copyright © 2017年 narita. All rights reserved.
//

import Foundation

// migemodoki
extension NSString
{
    // 漢字仮名交じり文字列をローマ字文字列に変換する
    func stringByTranslateJapaneseToRomaji() -> NSString
    {
        let theRomajiString = NSMutableString()
        
        let theTokenizer = CFStringTokenizerCreate(nil,
                                                   self,
                                                   CFRangeMake(0,self.length),
                                                   kCFStringTokenizerUnitWord,
                                                   CFLocaleCreate(kCFAllocatorDefault,
                                                                  CFLocaleIdentifier("Japanese" as CFString)))
        
        while(  [] != CFStringTokenizerAdvanceToNextToken(theTokenizer) )
        {
            let theCFType =  CFStringTokenizerCopyCurrentTokenAttribute(theTokenizer,
                                                                        kCFStringTokenizerAttributeLatinTranscription)
            
            theRomajiString.append(theCFType as! String)
        }
        
        return theRomajiString
    }
    
    // ローマ字を平仮名に変換する。これは、入力途中のローマ字を適当に平仮名に変えてしまうので、検索には使えない。
    // hを入力すると"は"や"お"が出力されてしまう。"h"はhのママで"は|ひ|ふ|へ|ほ"と展開したいので、別の方法を考える事
    func stringByTranslateRomajiToHiragana() -> NSString?
    {
        let theString: NSMutableString = self.mutableCopy() as! NSMutableString
        
        let theResult: Bool = CFStringTransform(theString, nil, kCFStringTransformLatinHiragana, false);

        if theResult
        {
            return theString.copy() as? NSString
        }
        else
        {
            return nil
        }
    }
    
    // コレジャナイ。 1文字しか変換できない
    @objc func incrimentalRomajiToHiragana() -> NSSet // Set<String>
    {
        let theSet = MINTLIncrimentalRomajiToHiragana.shared.transformToHiragana(toRomaji: self as String)
        
        return theSet as NSSet
    }

}

// 変換規則は、"/System/Library/Input Methods/JapaneseIM.app/Contents/PlugIns/JapaneseIM.appex/Contents/Resources/RomajiRule_Default.txt"の
// データを参考にした
// cat RomajiRule_Default.txt | awk '{print $1, $2}' | grep -v '^\s*#' | grep -v '^\s*$' > RomajiRule.txt

@objc class MINTLIncrimentalRomajiToHiragana: NSObject
{
    // シングルトンにしておく
    
    // MARK: - class methods
    
    static let shared: MINTLIncrimentalRomajiToHiragana =
    {
        return MINTLIncrimentalRomajiToHiragana()
    }()

    // MARK: - init methods
    
    override private init()
    {
        super.init()
    }

    // MARK: - dealloc
    
    // MARK: - NSCopying, hash, isEqual:
    
    // MARK: - NSCoding methods
    
    // MARK: - restorableState methods
    
    // MARK: - life cycle methods
    
    // MARK: - action methods
        
    // 文字列の組み合わせのSetを返す。
    func transformToHiraganaStrings(input: String) -> Set<String>
    {
        let theTuple = self.transformToHiraganas(toRomajis: input)
        
        if theTuple.l.count == 0 // 末尾の曖昧の数が0
        {
            return [theTuple.r]
        }
        else  // 末尾の曖昧の数が1以上
        {
            var theResult: Set<String> = []
            
            for i in theTuple.l
            {
                var theString: String =  String(theTuple.r)
                
                theString.append(i)
                
                theResult.insert(theString)
            }
            
            return theResult
        }
    }
    
    
    // 任意の文字列を変換してゆく
    // 先頭から順番に置き換えてゆく。
    // 先頭からの最大一致文字列で置き換えてゆく。
    // 曖昧で置き換えられない、アルファベとは残して次の文字へゆく
    // 末尾のアルファベットだけは、未入力の文字があるとみなして、複数の結果を返すようにする

    // rは末尾以外の文字列。lは可能性のある末尾の文字列
    func transformToHiraganas(toRomajis: String) -> ( r: String, l: Set<String>)
    {
        // 内部で使うだけのタプル
        // (preTransStr: String, postTransStrArry: Set<String>)
        typealias prePostPair = (preTransStr: String, postTransStrArry: Set<String>)
        
        // 内部で使うだけの関数
        // restに文字列を入れると、変換候補の配列を返す。再帰的な処理を行うために、引数と返値を同じにしている
        func f1(result: Array<prePostPair>, rest: String) -> (result: Array<prePostPair>, rest: String)
        {
            // 終了条件
            let theLength = rest.count

            if theLength == 0
            {
                return (result: result, rest: "")
            }

            // restが直接変換できるか確認。
            let theStringSet = transformToHiragana(toRomaji: rest)

            if theStringSet.isEmpty // 直接変換できないならば、restを末尾とそれ以外に分ける。
            {
                if theLength == 1 // 変換も分割もできない場合はその文字を結果とする
                {
                    return (result: result + [prePostPair(preTransStr: rest, postTransStrArry: [rest])] , rest: "")
                }
                
                
                // 末尾1文字とそれ以外の文字列に分ける。例："xxxy" =>("xxx", "y")
                let theLast = rest.last // 末尾文字
                let theRest = rest.dropLast()// 末尾以外全部

                // 末尾を足して結果とする
                // 再帰の返値のrest+string.lastが自分の返値のrestになる。
                let ret = f1(result: result, rest: String(theRest))

                return (result: ret.result , rest: ret.rest + String(theLast!))
            }
            else
            {
                let thePair = prePostPair(preTransStr: rest, postTransStrArry: theStringSet)
                
                return (result: result + [thePair] , rest:"")
            }
        }
        
        func f2(rest: String) -> Array<prePostPair>
        {
            var theRest = rest
            var theResult: Array<prePostPair> = []
            
            repeat
            {
                (result: theResult, rest: theRest) = f1(result: theResult, rest: theRest)
                
            }while( theRest != "")

            return theResult
        }
        
        let theResult = f2(rest: toRomajis)
        
        // 返値を構築し直す => ( r: String, l: Set<String>) rは末尾以外の文字列。lは可能性のある末尾の文字列
        // 末尾以外の文字の決定と、末尾文字の決定方法は異なる。
        let theLength = theResult.count
        
        var r: String = ""
        var l: Set<String> = []
        
        for i in 0 ..< theLength
        {
            let theResultItem = theResult[i]
            let thePosts = theResultItem.postTransStrArry

            if i != theLength - 1  // 末尾以外
            {
                if thePosts.count == 1
                {
                    r = r + thePosts.randomElement()!
                }
                else
                {
                    r = r + theResultItem.preTransStr
                }
            }
            else // 末尾
            {
                let thePre = theResultItem.preTransStr

                if thePosts.count == 1
                {
                    r = r + thePosts.randomElement()!
                }
                else if thePosts.count == 0
                {
                    r = r + thePre
                }
                else
                {
                    l.insert(thePre)
                    l.formUnion(thePosts)
                }
            }
        }
        
        return ( r: r, l: l)
    }
    
    // 1文字分しか変換できない
    func transformToHiragana(toRomaji: String) -> Set<String>
    {
        if let theRecode = transformArray.first(where:{return $0.romaji == toRomaji })
        {
            return theRecode.hiraganas
        }
        else
        {
            return Set<String>() // 変換できない場合は、空のSetを返す
        }
    }
    
    // MARK: - event handling methods
    
    // MARK: - drawing methods
    
    // MARK: - delegate/datasource methods
    
    // MARK: - accessor methods (in pairs)
    
    // 返値はArray<(romaji: String, hiraganas: set<String>)>
    
    let transformArray: Array<(romaji: String, hiraganas: Set<String>)> =
    {
        let theURL = Bundle.main.url(forResource: "RomajiRule", withExtension:"txt")
        
        let theRomajiRuleTxt = try! NSString(contentsOf: theURL!,
                                             encoding: String.Encoding.utf8.rawValue)
        
        let theLines = theRomajiRuleTxt.components(separatedBy: CharacterSet.newlines)
        
        let theTransformArray: Array<(romaji: String, hiraganas: Set<String>)> =
            theLines
                .filter()// "#"で始まる行をコメント行として除外
                {
                    $0.hasPrefix("#") == false
                }
                .compactMap()
                {
                    let theSeparetedLine = $0.components(separatedBy: CharacterSet.whitespaces)
                        
                    return ( romaji: theSeparetedLine.first!,
                         hiraganas: Set<String>(theSeparetedLine.dropFirst()))
                }
        
        return theTransformArray
    }()
    
    // MARK: - Utility methods

}
